@@ -21,13 +21,11 @@ class ConstrainedSR3(SR3):
         0.5\\|y-Xw\\|^2_2 + \\lambda R(u)
         + (0.5 / \\nu)\\|w-u\\|^2_2
 
-        \\text{subject to}
-
     .. math::
 
-        CW = d
+        \\text{subject to } Cw = d
 
-    over U and W, where :math:`R(U)` is a regularization function, C is a
+    over u and w, where :math:`R(u)` is a regularization function, C is a
     constraint matrix, and d is a vector of values. See the following
     reference for more details:
 
